\begin{code}
{-# OPTIONS --cubical #-}
module _ where

open import Agda.Primitive

open import Agda.Builtin.Cubical.Path
  renaming ( PathP to PathP_temp ; _≡_ to HiddenHomogeneousPathType)

open import Agda.Primitive.Cubical public
  renaming ( primIMin       to _∧_  -- I → I → I
           ; primIMax       to _∨_  -- I → I → I
           ; primINeg       to ~_   -- I → I
           -- ; isOneEmpty     to empty
           ; primComp       to compCCHM  -- This should not be used
           -- ; primHComp      to hcomp
           ; primTransp     to transp
           ; itIsOne        to 1=1 )

open import Agda.Builtin.Cubical.Sub
    renaming ( inc to primSubInc )
open import Agda.Builtin.Cubical.Glue
    using ( primGlue )

open import Agda.Builtin.Sigma public
open import Agda.Builtin.Bool public

infixr 30 _∙_

infixr 9 _∘_
_∘_ : ∀ {a b c}
        {A : Set a} {B : A → Set b} {C : {x : A} → B x → Set c} →
        (∀ {x} (y : B x) → C y) → (g : (x : A) → B x) →
        ((x : A) → C (g x))
f ∘ g = λ x → f (g x)

infix 2 Σ-syntax

Σ-syntax : ∀ {ℓ ℓ'} (A : Set ℓ) (B : A → Set ℓ') → Set (ℓ ⊔ ℓ')
Σ-syntax  = Σ

syntax Σ-syntax A (λ x → B)  = Σ[ x ∈ A ] B

_×_ : ∀ {ℓ ℓ'} (A : Set ℓ) (B : Set ℓ') → Set (ℓ ⊔ ℓ')
A × B = Σ A (λ _ → B)

infixr 5 _×_

private
  variable
    ℓ : Level
    A B C : Set ℓ
\end{code}

%<*PathP>
\begin{code}
PathP : (A : I → Set ℓ) → A i0 → A i1 → Set ℓ
\end{code}
%</PathP>
\begin{code}
PathP  = PathP_temp
\end{code}

%<*Path>
\begin{code}
_≡_ : {A : Set ℓ} → A → A → Set ℓ
_≡_ {A = A} x y = PathP (λ _ → A) x y
\end{code}
%</Path>

%<*refl>
\begin{code}
refl : {x : A} → x ≡ x
refl {x = x} = λ i → x
\end{code}
%</refl>

%<*Square>
\begin{code}
Square : {A : Set ℓ} {x0 x1 y0 y1 : A} →
           x0 ≡ x1 → y0 ≡ y1 → x0 ≡ y0 → x1 ≡ y1 → Set ℓ
Square p q r s  = PathP (λ i → p i ≡ q i) r s
\end{code}
%</Square>

%<*cong>
\begin{code}
cong : ∀ {B : A → Set ℓ} (f : (a : A) → B a) {x y} (p : x ≡ y) →
       PathP (λ i → B (p i)) (f x) (f y)
cong f p i  = f (p i)

congId : ∀ {x y : A} (p : x ≡ y) → cong (λ a → a) p ≡ p
congId p  = refl

congComp : ∀ (f : A → B) (g : B → C) {x y} (p : x ≡ y) →
           cong (g ∘ f ) p ≡ cong g (cong f p)
congComp f g p  = refl
\end{code}
%</cong>

%<*Sigmaeq>
\begin{code}
Σ-eq₂ : ∀ {A : Set} {B : A → Set} {p q : Σ[ x ∈ A ] (B x)} → (e : p ≡ q) →
          PathP (λ i → B (e i .fst)) (p .snd) (q .snd)
Σ-eq₂ = cong snd
\end{code}
%</Sigmaeq>

%<*sym>
\begin{code}
sym : {x y : A} → x ≡ y → y ≡ x
sym p  = λ i → p (~ i)
\end{code}
%</sym>

%<*symeqns>
\begin{code}
symInv : {x y : A} (p : x ≡ y) → sym (sym p) ≡ p
symInv p  = refl
\end{code}
%</symeqns>

%<*funext>
\begin{code}
funExt : {f g : A → B} → ((x : A) → f x ≡ g x) → f ≡ g
funExt p i x  = p x i
\end{code}
%</funext>

%<*transport>
\begin{code}
transport : A ≡ B → A → B
transport p a  = transp (λ i → p i) i0 a
\end{code}
%</transport>
~
%<*subst>
\begin{code}
subst : (B : A → Set ℓ) {x y : A} → x ≡ y → B x → B y
subst B p b  = transport (λ i → B (p i)) b
\end{code}
%</subst>

%<*J>
\begin{code}
J : ∀ {x : A} (P : ∀ y → x ≡ y → Set ℓ)
      (d : P x refl) {y : A} (p : x ≡ y)
      → P y p
J P d p  = transport (λ i → P (p i) (λ j → p (i ∧ j))) d
\end{code}
%</J>

%<*Jrefl>
\begin{code}
transportRefl : (x : A) → transport refl x ≡ x
transportRefl {A  = A} x i  = transp (λ _ → A) i x

JRefl : {x : A} (P : ∀ y → x ≡ y → Set ℓ)
         (d : P x refl) → J P d refl ≡ d
JRefl P d  = transportRefl d
\end{code}
%</Jrefl>

%<*partialBool>
\begin{code}
-- partialBool : ∀ i → Partial (i ∨ ~ i) Bool
-- partialBool i (i  = i0)  = true
-- partialBool i (i  = i1)  = false
\end{code}
%</partialBool>

%<*partialBool'>
\begin{code}
partialBool : ∀ i → Partial (i ∨ ~ i) Bool
partialBool i = λ { (i  = i0)  → true ; (i  = i1)  → false }
\end{code}
%</partialBool'>

%<*partialBool''>
\begin{code}
partialBool'' : ∀ i j → Partial (~ i ∨ i ∨ (i ∧ j)) Bool
partialBool'' i j  = λ  { (i = i1)           → true
                        ; (i = i1) (j = i1)  → true
                        ; (i = i0)           → false }
\end{code}
%</partialBool''>

%<*empty>
\begin{code}
empty : Partial i0 A
empty = λ { () }
\end{code}
%</empty>

%<*sub>
\begin{code}
_[_↦_] : (A : Set ℓ) (r : I) (u : Partial r A) → SSet ℓ
\end{code}
%</sub>
\begin{code}
A [ r ↦ u ]  = Sub A r u
\end{code}

%<*inc>
\begin{code}
inS : {r : I} (a : A) → A [ r ↦ (λ _ → a) ]
\end{code}
%</inc>
\begin{code}
inS = primSubInc
\end{code}

%<*ouc>
\begin{code}
outS : {r : I} {u : Partial r A} → A [ r ↦ u ] → A
\end{code}
%</ouc>
\begin{code}
outS = primSubOut
\end{code}

%<*transportFun>
\begin{code}
transportFun : (A B : I → Set) → (A i0 → B i0) → (A i1 → B i1)
transportFun A B f = transport (λ i → B i) ∘ f ∘ transport (λ i → A (~ i))
\end{code}
%</transportFun>

\begin{code}
transportFunOld : (A B : I → Set) → (A i0 → B i0) → (A i1 → B i1)
transportFunOld A B f = λ (x : A i1) → transport (λ i → B i) (f (transport (λ i → A (~ i)) x))
\end{code}

%<*transportPi>
\begin{code}
transportPi  : (A : I → Set) (B : (i : I) → A i → Set)
             → ((x : A i0) → B i0 x)
             → ((x : A i1) → B i1 x)
transportPi A B f = λ (x : A i1) → transport (λ j → B j (transp (λ i → A (j ∨ ~ i)) j x))
                                             (f (transport (λ i → A (~ i)) x))
\end{code}
%</transportPi>

%<*transportFunEq>
\begin{code}
transportFunEq : (A B : I → Set) → (f : A i0 → B i0) →
                 transportFun A B f ≡ transport (λ i → A i → B i) f
transportFunEq A B f = refl
\end{code}
%</transportFunEq>

\begin{code}
transportPiEq : (A : I → Set) (B : (i : I) → A i → Set) → (f : (x : A i0) → B i0 x) →
                transportPi A B f ≡ transport (λ i → (x : A i) → B i x) f
transportPiEq A B f = refl
\end{code}

%<*Sum>
\begin{code}
data _⊎_ (A B : Set) : Set where
  inl : A → A ⊎ B
  inr : B → A ⊎ B
\end{code}
%</Sum>

%<*transportSum>
\begin{code}
transportSum : (A B : I → Set) → A i0 ⊎ B i0 → A i1 ⊎ B i1
transportSum A B (inl x) = inl (transport (λ i → A i) x)
transportSum A B (inr x) = inr (transport (λ i → B i) x)
\end{code}
%</transportSum>

%<*hcomp>
\begin{code}
hcomp : {r : I} (u : I → Partial r A) (u₀ : A [ r ↦ u i0 ]) → A
\end{code}
%</hcomp>
\begin{code}
hcomp {r = r} u u₀ = primHComp {φ = r} (λ i → λ { (r = i1) → u i 1=1} ) (outS u₀)
\end{code}

%<*compPath>
\begin{code}
_∙_ : {x y z : A} → x ≡ y → y ≡ z → x ≡ z
_∙_ {x = x} p q i = hcomp (λ j → λ { (i = i0) → x ; (i = i1) → q j }) (inS (p i))
\end{code}
%</compPath>

%<*hfill>
\begin{code}
hfill : {r : I} (u : I → Partial r A) (u₀ : A [ r ↦ u i0 ]) → I → A
hfill {r = r} u u₀ i =
  hcomp (λ j → λ { (r = i1) → u (i ∧ j) 1=1 ; (i = i0) → outS u₀ }) (inS (outS u₀))
\end{code}
%</hfill>

\begin{code}
hfilleq0 : {r : I} (u : I → Partial r A) (u₀ : A [ r ↦ u i0 ]) → hfill u u₀ i0 ≡ outS u₀
hfilleq0 u u₀ = refl

hfilleq1 : {r : I} (u : I → Partial r A) (u₀ : A [ r ↦ u i0 ]) → hfill u u₀ i1 ≡ hcomp u u₀
hfilleq1 u u₀ = refl
\end{code}


%% \begin{code}
%% hfillTest : {r : I} (u : I → Partial r A) (u₀ : A [ r ↦ u i0 ]) → I → A
%% hfillTest {r = r} u u₀ i = {! hfill {r = r} u u₀ i1 !}
%% \end{code}
%%
%% %<*hfill'>
%% \begin{code}
%% hfill' : {r : I} (u : I → Partial r A) (u₀ : A [ r ↦ u i0 ]) → I → A
%% hfill' {r = r} u u₀ i = hcomp (λ j → λ { (r = i1) → u (i ∧ j) 1=1 ; (i = i0) → ouc u₀ }) u₀
%% \end{code}
%% %</hfill'>

%<*compPathRefl>
\begin{code}
compPathRefl : {x y : A} (p : x ≡ y) → p ∙ refl ≡ p
compPathRefl {x = x} {y = y} p j i =
  hfill (λ _ → λ { (i = i0) → x ; (i = i1) → y }) (inS (p i)) (~ j)
\end{code}
%</compPathRefl>

-- Unfolded version
transportPath : (A : I → Set) (x y : (i : I) → A i) → x i0 ≡ y i0 → x i1 ≡ y i1
transportPath A x y p =
  λ i → hcomp (λ j → λ { (i = i0) → transp (λ k → A (j ∨ k)) j (x j)
                       ; (i = i1) → transp (λ k → A (j ∨ k)) j (y j)})
              (inc (transport (λ j → A j) (p i)))

%<*comp>
\begin{code}
comp : (A : I → Set ℓ) {r : I} (u : (i : I) → Partial r (A i)) (u0 : A i0 [ r ↦ u i0 ]) → A i1
comp A {r = r} u u0 =
  hcomp (λ i → λ { (r = i1) → transp (λ j → A (i ∨ j)) i (u _ 1=1) })
        (inS (transport (λ i → A i) (outS u0)))
\end{code}
%</comp>

%<*transportPath>
\begin{code}
transportPath : (A : I → Set) (x y : (i : I) → A i) → x i0 ≡ y i0 → x i1 ≡ y i1
transportPath A x y p =
  λ i → comp A (λ j → λ { (i = i0) → x j ; (i = i1) → y j}) (inS (p i))
\end{code}
%</transportPath>

%<*transportPathBad>
transportPathBad : (A : I → Set) (x y : (i : I) → A i) → x i0 ≡ y i0 → x i1 ≡ y i1
transportPathBad A x y p = λ i → transport (λ j → A j) (p i)
%</transportPathBad>


%% Experiment with nested pathes


\begin{code}
sym2 : ∀{a b : A} {p q : a ≡ b} (P : p ≡ q) → sym q ≡ sym p
sym2 P i j = P (~ i) (~ j)

-- sym P @(sym q,sym p) i @b,a j = P @p,q (~ i) @a,b (~ j)
-- Boundary checks:
-- i = i0:  sym q @b,a j  = q @a,b (~ j)
--          =  (λ j → q @a,b (~ j)) @b,a j
-- j = j0:  b = b
\end{code}

%<*equiv>
\begin{code}
fiber : {A B : Set ℓ} (f : A → B) (y : B) → Set ℓ
fiber {A = A} f y = Σ[ x ∈ A ] f x ≡ y

isContr : Set ℓ → Set ℓ
isContr A = Σ[ x ∈ A ] (∀ y → x ≡ y)

record isEquiv {ℓ} {A B : Set ℓ} (f : A → B) : Set ℓ where
  field
    equiv-proof : (y : B) → isContr (fiber f y)

_≃_ : (A B : Set ℓ) → Set ℓ
A ≃ B = Σ[ f ∈ (A → B) ] (isEquiv f)
\end{code}
%</equiv>

\begin{code}
open isEquiv
\end{code}

%<*idequiv>
\begin{code}
idfun : (A : Set ℓ) → A → A
idfun _ x = x

idIsEquiv : (A : Set ℓ) → isEquiv (idfun A)
equiv-proof (idIsEquiv A) y = ((y , refl) , λ z i → z .snd (~ i) , λ j → z .snd (~ i ∨ j))

idEquiv : (A : Set ℓ) → A ≃ A
idEquiv A = (idfun A , idIsEquiv A)
\end{code}
%</idequiv>

%<*ghcomp>
\begin{code}
ghcomp : {r : I} (u : I → Partial r A) (u₀ : A [ r ↦ u i0 ]) → A
ghcomp {r = r} u u₀ =
  hcomp (λ j → λ { (r = i1) → u j 1=1 ; (r = i0) → outS u₀ }) (inS (outS u₀))
\end{code}
%</ghcomp>

