\begin{code}
{-# OPTIONS --cubical --safe #-}
module _ where

open import Cubical.Foundations.Prelude public
  hiding ( _≡_ ; _≡⟨_⟩_ ; _∎ ; isPropIsContr)
  renaming (J   to JPath
           ; JRefl     to JPathRefl
           ; sym       to symPath
           ; _∙_       to compPath
           ; cong      to congPath
           ; subst     to substPath
           ; funExt    to funExtPath
           ; isContr   to isContrPath
           ; isProp    to isPropPath )
open import Cubical.Foundations.Isomorphism
open import Cubical.Foundations.Equiv
  renaming ( fiber     to fiberPath
           ; isEquiv   to isEquivPath
           ; equivFun  to equivFunPath
           ; isPropIsEquiv to isPropIsEquivPath )
  hiding   ( equivCtr
           ; equivIsEquiv )
open import Cubical.Foundations.Univalence hiding (univalence)
open import Cubical.Data.Nat using (ℕ)

private
  variable
    ℓ ℓ' : Level

_≡_ : ∀ {ℓ} {A : Set ℓ} → A → A → Set ℓ
_≡_ {A = A} = PathP (λ i → A)

\end{code}

%<*eq>
\begin{code}
data Eq {A : Set ℓ} (x : A) : A → Set ℓ where
  reflEq : Eq x x
\end{code}
%</eq>

%<*eqdefs>
\begin{code}
ap : {A B : Set ℓ} (f : A → B) {x y : A} → Eq x y → Eq (f x) (f y)
ap f reflEq = reflEq
\end{code}
%</eqdefs>

%<*pathToEq>
\begin{code}
eqToPath : {A : Set ℓ} {x y : A} → Eq x y → x ≡ y
eqToPath reflEq = refl

pathToEq : {A : Set ℓ} {x y : A} → x ≡ y → Eq x y
pathToEq {x = x} p = transport (λ i → Eq x (p i)) reflEq
\end{code}
%</pathToEq>

\begin{code}
eqToPath-pathToEq : {A : Set ℓ} {x y : A} → (p : x ≡ y) → eqToPath (pathToEq p) ≡ p
eqToPath-pathToEq p =
  JPath (λ _ h → Path _ (eqToPath (pathToEq h)) h)
        (congPath eqToPath (transportRefl reflEq)) p

pathToEq-eqToPath : {A : Set ℓ} {x y : A} → (p : Eq x y) → pathToEq (eqToPath p) ≡ p
pathToEq-eqToPath reflEq = transportRefl reflEq

Path≃Eq : {A : Set ℓ} {x y : A} → (x ≡ y) ≃ (Eq x y)
Path≃Eq = isoToEquiv (iso pathToEq eqToPath pathToEq-eqToPath eqToPath-pathToEq)

PathEqEq : {A : Set ℓ} {x y : A} → Eq (x ≡ y) (Eq x y)
PathEqEq = pathToEq (ua Path≃Eq)
\end{code}

%<*patheq>
\begin{code}
Path≡Eq : {A : Set ℓ} {x y : A} → (x ≡ y) ≡ (Eq x y)
Path≡Eq = ua Path≃Eq
\end{code}
%</patheq>

%<*funexteq>
\begin{code}
funExtEq : {A : Set ℓ} {B : A → Set ℓ'} {f g : (x : A) → B x} →
           ((x : A) → Eq (f x) (g x)) → Eq f g
funExtEq p = pathToEq (λ i x → eqToPath (p x) i)
\end{code}
%</funexteq>

%<*sym0>
\begin{code}
sym0 : (x : ℕ) → Eq x 0 → Eq 0 x
sym0 .0 reflEq = reflEq
\end{code}
%</sym0>
