\begin{code}
{-# OPTIONS --cubical --safe #-}
module _ where

open import Agda.Primitive

open import Agda.Builtin.Cubical.Path

open import Agda.Primitive.Cubical public
  renaming ( primIMin       to _∧_  -- I → I → I
           ; primIMax       to _∨_  -- I → I → I
           ; primINeg       to ~_   -- I → I
           ; isOneEmpty     to empty
           ; primComp       to compCCHM  -- This should not be used
           ; primHComp      to hcomp
           ; primTransp     to transp
           ; itIsOne        to 1=1 )

open import Agda.Builtin.Cubical.Sub

open import Agda.Builtin.Cubical.Glue public
  using ( isEquiv       -- ∀ {ℓ ℓ'} {A : Set ℓ} {B : Set ℓ'} (f : A → B) → Set (ℓ ⊔ ℓ')

        ; equiv-proof   -- ∀ (y : B) → isContr (fiber f y)

        ; _≃_           -- ∀ {ℓ ℓ'} (A : Set ℓ) (B : Set ℓ') → Set (ℓ ⊔ ℓ')

        ; equivFun      -- ∀ {ℓ ℓ'} {A : Set ℓ} {B : Set ℓ'} → A ≃ B → A → B

        ; equivProof    -- ∀ {ℓ ℓ'} (T : Set ℓ) (A : Set ℓ') (w : T ≃ A) (a : A) φ →
                        -- Partial φ (fiber (equivFun w) a) → fiber (equivFun w) a

        ; primGlue      -- ∀ {ℓ ℓ'} (A : Set ℓ) {φ : I} (T : Partial φ (Set ℓ'))
                        -- → (e : PartialP φ (λ o → T o ≃ A)) → Set ℓ'

        ; prim^unglue   -- ∀ {ℓ ℓ'} {A : Set ℓ} {φ : I} {T : Partial φ (Set ℓ')}
                        -- → {e : PartialP φ (λ o → T o ≃ A)} → primGlue A T e → A
        ; prim^glue
        -- The ∀ operation on I. This is commented out as it is not currently used for anything
        -- ; primFaceForall -- (I → I) → I
        )
  renaming ( pathToEquiv to lineToEquiv  -- ∀ {ℓ : I → Level} (P : (i : I) → Set (ℓ i)) → P i0 ≃ P i1
           )
open import Agda.Builtin.Sigma public
open import Agda.Builtin.Bool public

infix 2 Σ-syntax

Σ-syntax : ∀ {ℓ ℓ'} (A : Set ℓ) (B : A → Set ℓ') → Set (ℓ ⊔ ℓ')
Σ-syntax  = Σ

syntax Σ-syntax A (λ x → B)  = Σ[ x ∈ A ] B

_×_ : ∀ {ℓ ℓ'} (A : Set ℓ) (B : Set ℓ') → Set (ℓ ⊔ ℓ')
A × B = Σ A (λ _ → B)

infixr 5 _×_

refl : ∀ {ℓ} {A : Set ℓ} {x : A} → x ≡ x
refl {x = x} = λ i → x

_[_↦_] : ∀ {ℓ} (A : Set ℓ) (r : I) (u : Partial r A) → SSet ℓ
A [ r ↦ u ]  = Sub A r u

ouc : ∀ {ℓ} {A : Set ℓ} {r : I} {u : Partial r A} → A [ r ↦ u ] → A
ouc = primSubOut

idfun : ∀ {ℓ} → (A : Set ℓ) → A → A
idfun _ x = x

idIsEquiv : ∀ {ℓ} (A : Set ℓ) → isEquiv (idfun A)
equiv-proof (idIsEquiv A) y = ((y , refl) , λ z i → z .snd (~ i) , λ j → z .snd (~ i ∨ j))

idEquiv : ∀ {ℓ} (A : Set ℓ) → A ≃ A
idEquiv A = (idfun A , idIsEquiv A)

transport : ∀ {ℓ} {A B : Set ℓ} → A ≡ B → A → B
transport p a  = transp (λ i → p i) i0 a

variable
  ℓ : Level
  A : Set ℓ
\end{code}

%<*Glue>
\begin{code}
Glue : (B : Set ℓ) {r : I} → Partial r (Σ[ A ∈ Set ℓ ] (A ≃ B)) → Set ℓ
\end{code}
%</Glue>
\begin{code}
Glue B Ae = primGlue B (λ x → Ae x .fst) (λ x → Ae x .snd)
\end{code}

%<*glue>
\begin{code}
glue    :  {B : Set ℓ} {r : I} {Ae :  Partial r (Σ[ A ∈ Set ℓ ] (A ≃ B))} →
           (t : PartialP r (λ o → Ae o .fst)) → B [ r ↦ (λ x → Ae x .snd .fst (t x)) ] → Glue B Ae

unglue  :  {B : Set ℓ} (r : I) {Ae :  Partial r (Σ[ A ∈ Set ℓ ] (A ≃ B))} → Glue B Ae → B
\end{code}
%</glue>

\begin{code}
glue x a = prim^glue x (ouc a)
unglue r = prim^unglue {φ = r}
\end{code}

%<*ua>
\begin{code}
ua : {A B : Set ℓ} → A ≃ B → A ≡ B
ua {A = A} {B = B} e i = Glue B (λ { (i = i0) → (A , e) ; (i = i1) → (B , idEquiv B) })
\end{code}
%</ua>

%<*uabeta>
\begin{code}
uaβ : {A B : Set ℓ} (e : A ≃ B) (a : A) → transport (ua e) a ≡ e .fst a
uaβ {B = B} e a = λ i → transp (λ _ → B) i (e .fst a)
\end{code}
%</uabeta>
