\begin{code}
{-# OPTIONS --cubical --safe #-}
module _ where

open import Cubical.Core.Everything

open import Cubical.Foundations.Prelude
open import Cubical.Foundations.Function
open import Cubical.Foundations.Equiv
open import Cubical.Foundations.Isomorphism
open import Cubical.Foundations.Transport

open import Cubical.Data.Nat
\end{code}

%<*int>
\begin{code}
data ℤ : Set where
  pos : (n : ℕ) → ℤ
  neg : (n : ℕ) → ℤ
  posneg : pos 0 ≡ neg 0
\end{code}
%</int>

%<*sucint>
\begin{code}
sucℤ : ℤ → ℤ
sucℤ (pos n)        = pos (suc n)
sucℤ (neg zero)     = pos 1
sucℤ (neg (suc n))  = neg n
sucℤ (posneg i)     = pos 1
\end{code}
%</sucint>

\begin{code}
predℤ : ℤ → ℤ
predℤ (pos zero)     = neg 1
predℤ (pos (suc n))  = pos n
predℤ (neg n)        = neg (suc n)
predℤ (posneg _)     = neg 1

sucPredℤ : (n : ℤ) → sucℤ (predℤ n) ≡ n
sucPredℤ (pos zero)     = sym posneg
sucPredℤ (pos (suc _))  = refl
sucPredℤ (neg _)        = refl
sucPredℤ (posneg i) j   = posneg (i ∨ ~ j)

predSucℤ : (n : ℤ) → predℤ (sucℤ n) ≡ n
predSucℤ (pos _)        = refl
predSucℤ (neg zero)     = posneg
predSucℤ (neg (suc _))  = refl
predSucℤ (posneg i) j   = posneg (i ∧ j)
\end{code}

%<*addint>
\begin{code}
_+ℤ_ : ℤ → ℤ → ℤ
m +ℤ (pos (suc n))  =  sucℤ   (m +ℤ pos n)
m +ℤ (neg (suc n))  =  predℤ  (m +ℤ neg n)
m +ℤ _              =  m
\end{code}
%</addint>

%<*sucpath>
\begin{code}
sucPathℤ : ℤ ≡ ℤ
sucPathℤ  = isoToPath (iso sucℤ predℤ sucPredℤ predSucℤ)
\end{code}
%</sucpath>

%<*addeq1>
\begin{code}
addEq : ℕ → ℤ ≡ ℤ
addEq zero     = refl
addEq (suc n)  = addEq n ∙ sucPathℤ
\end{code}
%</addeq1>

\begin{code}
predPathℤ : ℤ ≡ ℤ
predPathℤ  = isoToPath (iso predℤ sucℤ predSucℤ sucPredℤ)

subEq : ℕ → ℤ ≡ ℤ
subEq zero     = refl
subEq (suc n)  = subEq n ∙ predPathℤ
\end{code}

%<*addint'>
\begin{code}
addℤ : ℤ → ℤ → ℤ
addℤ m (pos n)     = transport (addEq n) m
addℤ m (neg n)     = transport (subEq n) m
addℤ m (posneg _)  = m
\end{code}
%</addint'>

%<*addequiv'>
\begin{code}
isEquivAddℤ : (m : ℤ) → isEquiv (λ n → addℤ n m)
isEquivAddℤ (pos n)     = isEquivTransport (addEq n)
isEquivAddℤ (neg n)     = isEquivTransport (subEq n)
isEquivAddℤ (posneg i)  = isEquivTransport refl
\end{code}
%</addequiv'>

%<*addeq>
\begin{code}
addℤ≡+ℤ : addℤ ≡ _+ℤ_
addℤ≡+ℤ i  m (pos (suc n))  = sucℤ   (addℤ≡+ℤ i m (pos n))
addℤ≡+ℤ i  m (neg (suc n))  = predℤ  (addℤ≡+ℤ i m (neg n))
addℤ≡+ℤ i  m (pos zero)     = m
addℤ≡+ℤ i  m (neg zero)     = m
addℤ≡+ℤ _  m (posneg _)     = m

isEquiv+ℤ : (m : ℤ) → isEquiv (λ n → n +ℤ m)
isEquiv+ℤ = subst (λ _+_ → (m : ℤ) → isEquiv (λ n → n + m)) addℤ≡+ℤ isEquivAddℤ
\end{code}
%</addeq>
