\begin{code}
{-# OPTIONS --cubical --safe #-}
module _ where

open import Cubical.Core.Everything
open import Cubical.Foundations.Everything
open import Cubical.Data.Nat
open import Cubical.Data.Unit
open import Cubical.Data.Prod hiding (intro)
open import Cubical.HITs.PropositionalTruncation as PT
open import Cubical.HITs.SetTruncation
open import HitInt

variable
  ℓ : Level

\end{code}

%<*setquot>
\begin{code}
data _/_ (A : Set ℓ) (R : A → A → Set ℓ) : Set ℓ where
  [_]    :  A → A / R
  eq     :  (a b : A) → (r : R a b) → [ a ] ≡ [ b ]
  trunc  :  (x y : A / R) → (p q : x ≡ y) → p ≡ q
\end{code}
%</setquot>
\begin{code}


elimEq :  {A : Set ℓ} {R : A → A → Set ℓ} {B : A / R → Set ℓ} →
          ((x : A / R ) → isProp (B x)) →
          {x y : A / R} →
          (eq : x ≡ y) →
          (bx : B x) →
          (by : B y) →
          PathP (λ i → B (eq i)) bx by
elimEq {A = A} {B = B} Bprop {x = x} =
  J (λ y eq → ∀ bx by → PathP (λ i → B (eq i)) bx by) (λ bx by → Bprop x bx by)


elimSetQuotientsProp : {A : Set ℓ} {R : A → A → Set ℓ} {B : A / R → Set ℓ} →
                       ((x : A / R ) → isProp (B x)) →
                       (f : (a : A) → B ( [ a ])) →
                       (x : A / R) → B x
elimSetQuotientsProp Bprop f [ x ] = f x
elimSetQuotientsProp Bprop f (trunc x y p q i j) =
  isOfHLevel→isOfHLevelDep 2 (λ x → isProp→isSet (Bprop x))
              (g x) (g y) (cong g p) (cong g q) (trunc x y p q) i j
    where
    g = elimSetQuotientsProp Bprop f
elimSetQuotientsProp Bprop f (eq a b r i) = elimEq Bprop (eq a b r) (f a) (f b) i

-- lemma 6.10.2 in hott book
-- TODO: defined truncated Sigma as ∃
[]surjective : {A : Set ℓ} {R : A → A → Set ℓ} (x : A / R) → ∥ Σ[ a ∈ A ] [ a ] ≡ x ∥
[]surjective = elimSetQuotientsProp (λ x → squash) (λ a → ∣ a , refl ∣)


elimSetQuotients : {A : Set ℓ} {R : A → A → Set ℓ} {B : A / R → Set ℓ} →
                   (Bset : (x : A / R) → isSet (B x)) →
                   (f : (a : A) → (B [ a ])) →
                   (feq : (a b : A) (r : R a b) →
                          PathP (λ i → B (eq a b r i)) (f a) (f b)) →
                   (x : A / R) → B x
elimSetQuotients Bset f feq [ a ] = f a
elimSetQuotients Bset f feq (eq a b r i) = feq a b r i
elimSetQuotients Bset f feq (trunc x y p q i j) =
   isOfHLevel→isOfHLevelDep 2 Bset
              (g x) (g y) (cong g p) (cong g q) (trunc x y p q) i j
    where
      g = elimSetQuotients Bset f feq
\end{code}
%<*setquotuniv>
\begin{code}
setQuotUniversal : {A B : Set ℓ} {R : A → A → Set ℓ} → isSet B →
                   (A / R → B) ≃ (Σ[ f ∈ (A → B) ] ((a b : A) → R a b → f a ≡ f b))
\end{code}
%</setquotuniv>
\begin{code}
setQuotUniversal Bset = isoToEquiv (iso intro'' elim'' elimRightInv elimLeftInv)
  where
  intro'' = λ g →  (λ a → g [ a ]) , λ a b r i → g (eq a b r i)
  elim'' = λ h → elimSetQuotients (λ x → Bset) (fst h) (snd h)

  elimRightInv : ∀ h → intro'' (elim'' h) ≡ h
  elimRightInv h = refl

  elimLeftInv : ∀ g → elim'' (intro'' g) ≡ g
  elimLeftInv = λ g → funExt (λ x → PT.elim {P = λ sur → elim'' (intro'' g) x ≡ g x}
    (λ sur → Bset (elim'' (intro'' g) x) (g x))
    (λ sur → (cong (elim'' (intro'' g)) (sym (snd sur))) ∙ (cong g (snd sur))) ([]surjective x)
    )


-- We can define int as follows

--   (x1,x2) ~ (y1,y2) := (x1 + y2 = x2 + y1)
natrel : ℕ × ℕ → ℕ × ℕ → Set
natrel (x₁ , x₂) (y₁ , y₂) = x₁ + y₂ ≡ x₂ + y₁

int : Set
int = (ℕ × ℕ) / natrel
\end{code}
