\begin{code}
{-# OPTIONS --cubical --safe --guardedness #-}
module Stream where

open import Cubical.Core.Everything

open import Cubical.Data.Nat
open import Cubical.Foundations.Equiv
open import Cubical.Foundations.Prelude
open import Cubical.Foundations.Isomorphism
open import Cubical.Foundations.Univalence
\end{code}

%<*Stream>
\begin{code}
record Stream (A : Set) : Set where
  coinductive; constructor _,_
  field
    head  : A
    tail  : Stream A
\end{code}
%</Stream>


\begin{code}
open Stream public

variable
  A B : Set
\end{code}

%<*mapS>
\begin{code}
mapS : (A → B) → Stream A → Stream B
mapS f xs .head  = f (xs .head)
mapS f xs .tail  = mapS f (xs .tail)
\end{code}
%</mapS>

%<*mapSid>
\begin{code}
mapS-id : (xs : Stream A) → mapS (λ x → x) xs ≡ xs
mapS-id xs i .head  = xs .head
mapS-id xs i .tail  = mapS-id (xs .tail) i
\end{code}
%</mapSid>

%<*elimS>
\begin{code}
Stream-η : (xs : Stream A) → xs ≡ (head xs , tail xs)
Stream-η xs i .head  = head xs
Stream-η xs i .tail  = tail xs

elimS : (P : Stream A → Set) (c : ∀ x xs → P (x , xs)) (xs : Stream A) → P xs
elimS P c xs = transport  (λ i → P (Stream-η xs (~ i)))
                          (c (xs .head) (xs .tail))
\end{code}
%</elimS>

\begin{code}
module Equality≅Bisimulation where
\end{code}

%<*bisimilarity>
\begin{code}
  record _≈_ (xs ys : Stream A) : Set where
    coinductive
    field
      ≈head  : xs .head  ≡ ys .head
      ≈tail  : xs .tail  ≈ ys .tail
\end{code}
%</bisimilarity>

\begin{code}
  open _≈_
\end{code}

%<*bisim>
\begin{code}
  bisim :  ∀ {xs ys : Stream A} → xs ≈ ys → xs ≡ ys
  bisim xs≈ys i .head  = xs≈ys .≈head i
  bisim xs≈ys i .tail  = bisim (xs≈ys .≈tail) i
\end{code}
%</bisim>

\begin{code}
  misib : ∀ {xs ys : Stream A} → xs ≡ ys → xs ≈ ys
  ≈head (misib p)  = λ i → head (p i)
  ≈tail (misib p)  = misib (λ i → tail (p i))
\end{code}

\begin{code}
  iso1 : ∀ {xs ys : Stream A} → (p : xs ≡ ys) → bisim (misib p) ≡ p
  head (iso1 p i j) = head (p j)
  tail (iso1 p i j) = iso1 (λ i → tail (p i)) i j

  iso2 : ∀ {xs ys : Stream A} → (p : xs ≈ ys) → misib (bisim p) ≡ p
  ≈head (iso2 p i) = ≈head p
  ≈tail (iso2 p i) = iso2 (≈tail p) i

  path≃bisim : ∀ {xs ys : Stream A} → (xs ≡ ys) ≃ (xs ≈ ys)
  path≃bisim = isoToEquiv (iso misib bisim iso2 iso1)
\end{code}

%<*pathbisim>
\begin{code}
  path≡bisim : ∀ {xs ys : Stream A} → (xs ≡ ys) ≡ (xs ≈ ys)
\end{code}
%</pathbisim>
\begin{code}
  path≡bisim = ua path≃bisim
\end{code}
