\begin{code}
{-# OPTIONS --cubical --safe #-}
module _ where

open import Cubical.Foundations.Prelude
open import Cubical.Foundations.Isomorphism
open import Cubical.Data.Prod
\end{code}

%<*s1>
\begin{code}
data S¹ : Set where
  base  : S¹
  loop  : base ≡ base
\end{code}
%</s1>
%<*torus>
\begin{code}
data Torus : Set where
  point   : Torus
  line1   : point ≡ point
  line2   : point ≡ point
  square  : PathP  (λ i  →  line1 i ≡  line1 i) line2 line2
\end{code}
%</torus>

%<*t2c>
\begin{code}
t2c : Torus → S¹ × S¹
t2c point         = (base    , base)
t2c (line1 i)     = (loop i  , base)
t2c (line2 j)     = (base    , loop j)
t2c (square i j)  = (loop i  , loop j)
\end{code}
%</t2c>

%<*c2t>
\begin{code}
c2t : S¹ × S¹ → Torus
c2t (base   , base)    = point
c2t (loop i , base)    = line1 i
c2t (base   , loop j)  = line2 j
c2t (loop i , loop j)  = square i j
\end{code}
%</c2t>

%<*c2tt2c>
\begin{code}
c2t-t2c : (t : Torus) → c2t (t2c t) ≡ t
c2t-t2c point         = refl
c2t-t2c (line1 _)     = refl
c2t-t2c (line2 _)     = refl
c2t-t2c (square _ _)  = refl
\end{code}
%</c2tt2c>

%<*t2cc2t>
\begin{code}
t2c-c2t : (p : S¹ × S¹) → t2c (c2t p) ≡ p
t2c-c2t (base    , base)    = refl
t2c-c2t (base    , loop _)  = refl
t2c-c2t (loop _  , base)    = refl
t2c-c2t (loop _  , loop _)  = refl
\end{code}
%</t2cc2t>

%<*teqs1s1>
\begin{code}
Torus≡S¹×S¹ : Torus ≡ S¹ × S¹
Torus≡S¹×S¹  = isoToPath (iso t2c c2t t2c-c2t c2t-t2c)
\end{code}
%</teqs1s1>
